import asyncio
import inspect
from functools import wraps
from typing import (
    Any,
    Dict,
    List,
    Tuple,
    Union,
    _UnionGenericAlias,
    get_args,
    get_origin,
    get_type_hints,
)

from typeguard import check_type
from typing_extensions import Annotated, Doc, _AnnotatedAlias

TYPE_TO_STRING = {
    int: "integer",
    str: "string",
    float: "number",
    bool: "boolean",
    Any: "any",
    List: "array",
    list: "array",
    dict: "object",
    Dict: "object",
}
FORMAT_TYPE_STRING = {
    "int": "integer",
    "str": "string",
    "float": "number",
    "bool": "boolean",
    "list": "array",
    "dict": "object",
}

TYPE_STRING_TO_TYPE = {
    "integer": int,
    "int": int,
    "string": str,
    "str": str,
    "float": float,
    "boolean": bool,
    "bool": bool,
    "any": Any,
    "array": list,
    "list": list,
    "object": dict,
}


def format_type_string(type_str: str) -> str:
    """Convert a type string to a standard format."""
    return FORMAT_TYPE_STRING.get(type_str, type_str)


def _is_typing(obj):
    from typing import _Final  # type: ignore

    return isinstance(obj, _Final)


def _is_instance_of_generic_type(obj, generic_type):
    """Check if an object is an instance of a generic type."""
    if generic_type is Any:
        return True  # Any type is compatible with any object

    origin = get_origin(generic_type)
    if origin is None:
        return isinstance(obj, generic_type)  # Handle non-generic types

    args = get_args(generic_type)
    if not args:
        return isinstance(obj, origin)

    # Check if object matches the generic origin (like list, dict)
    if not _is_typing(origin):
        return isinstance(obj, origin)

    objs = [obj for _ in range(len(args))]

    # For each item in the object, check if it matches the corresponding type argument
    for sub_obj, arg in zip(objs, args):
        # Skip check if the type argument is Any
        if arg is not Any:
            if _is_typing(arg):
                sub_args = get_args(arg)
                if (
                    sub_args
                    and not _is_typing(sub_args[0])
                    and not isinstance(sub_obj, sub_args[0])
                ):
                    return False
            elif not isinstance(sub_obj, arg):
                return False
    return True


def _check_type(obj, t) -> bool:
    try:
        check_type(obj, t)
        return True
    except Exception:
        return False


def _get_orders(obj, arg_types):
    try:
        orders = [i for i, t in enumerate(arg_types) if _check_type(obj, t)]
        return orders[0] if orders else int(1e8)
    except Exception:
        return int(1e8)


def _sort_args(func, args, kwargs):
    sig = inspect.signature(func)
    type_hints = get_type_hints(func)

    arg_types = [
        type_hints[param_name]
        for param_name in sig.parameters
        if param_name != "return" and param_name != "self"
    ]

    if "self" in sig.parameters:
        self_arg = [args[0]]
        other_args = args[1:]
    else:
        self_arg = []
        other_args = args

    sorted_args = sorted(
        other_args,
        key=lambda x: _get_orders(x, arg_types),
    )
    return (*self_arg, *sorted_args), kwargs


def rearrange_args_by_type(func):
    """Decorator to rearrange the arguments of a function by type.

    Examples:

        .. code-block:: python

            from dbgpt.util.function_utils import rearrange_args_by_type


            @rearrange_args_by_type
            def sync_regular_function(a: int, b: str, c: float):
                return a, b, c


            assert instance.sync_class_method(1, "b", 3.0) == (1, "b", 3.0)
            assert instance.sync_class_method("b", 3.0, 1) == (1, "b", 3.0)

    """

    @wraps(func)
    def sync_wrapper(*args, **kwargs):
        sorted_args, sorted_kwargs = _sort_args(func, args, kwargs)
        return func(*sorted_args, **sorted_kwargs)

    @wraps(func)
    async def async_wrapper(*args, **kwargs):
        sorted_args, sorted_kwargs = _sort_args(func, args, kwargs)
        return await func(*sorted_args, **sorted_kwargs)

    return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper


def type_to_string(obj: Any, default_type: str = "unknown") -> Tuple[str, List[str]]:
    """Convert a type to a string representation."""
    # Check NoneType
    if obj is type(None) or obj is None:
        return "null", []

    # Get the origin of the type
    origin = getattr(obj, "__origin__", None)
    if origin:
        if _is_typing(origin) and not isinstance(obj, _UnionGenericAlias):
            obj = origin
            origin = origin.__origin__
        elif _is_typing(origin) and isinstance(obj, _UnionGenericAlias):
            # Remove NoneType from Union
            return type_to_string(obj.__args__[0], default_type)
        # Handle special cases like List[int]
        if origin is Union and hasattr(obj, "__args__"):
            subtypes = ", ".join(
                type_to_string(t, default_type)[0]
                for t in obj.__args__
                if t is not type(None)
            )
            # return f"Optional[{subtypes}]"
            return subtypes, []
        elif origin is list or origin is List:
            subtypes = list(type_to_string(t, default_type)[0] for t in obj.__args__)
            # return f"List[{subtypes}]"
            return "array", subtypes
        elif origin in [dict, Dict]:
            # key_type, value_type = (type_to_string(t) for t in obj.__args__)
            # subtypes = ", ".join(type_to_string(t) for t in obj.__args__)
            # return f"Dict[{key_type}, {value_type}]"
            return "object", []
        # Handle tuple
        elif origin is tuple:
            subtypes = list(type_to_string(t, default_type)[0] for t in obj.__args__)
            return "array", subtypes
        elif origin is Annotated:
            subtypes = list(type_to_string(t, default_type)[0] for t in obj.__args__)
            return subtypes[0], []
        elif origin is _UnionGenericAlias:
            subtypes = list(type_to_string(t, default_type)[0] for t in obj.__args__)
            return subtypes, []
    elif obj is list:
        return "array", []
    elif obj is dict:
        return "object", []
    elif obj is tuple:
        return "array", []
    elif hasattr(obj, "__args__"):
        subtypes = ", ".join(
            type_to_string(t, default_type)[0]
            for t in obj.__args__
            if t is not type(None)
        )
        return subtypes, []

    if obj in TYPE_TO_STRING:
        return TYPE_TO_STRING[obj], []

    if default_type == "unknown" and hasattr(obj, "__name__"):
        return obj.__name__, []
    return default_type, []


def parse_param_description(name: str, obj: Any) -> str:
    default_type_title = name.replace("_", " ").title()
    if isinstance(obj, _AnnotatedAlias):
        metadata = obj.__metadata__
        docs = [arg for arg in metadata if isinstance(arg, Doc)]
        doc_str = docs[0].documentation if docs else default_type_title
    else:
        doc_str = default_type_title
    return doc_str



import logging
import os
from concurrent.futures import Executor
from dataclasses import dataclass, field
from typing import Iterator, Optional, Type, Union

from dbgpt.core import MessageConverter, ModelMetadata, ModelOutput, ModelRequest
from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.core.interface.parameter import LLMDeployModelParameters
from dbgpt.model.proxy.base import (
    AsyncGenerateStreamFunction,
    GenerateStreamFunction,
    ProxyLLMClient,
    register_proxy_model_adapter,
)
from dbgpt.model.proxy.llms.proxy_model import ProxyModel, parse_model_request
from dbgpt.util.i18n_utils import _

from .chatgpt import OpenAICompatibleDeployModelParameters

_DEFAULT_MODEL = "glm-4-plus"

logger = logging.getLogger(__name__)


@auto_register_resource(
    label=_("Zhipu Proxy LLM"),
    category=ResourceCategory.LLM_CLIENT,
    tags={"order": TAGS_ORDER_HIGH},
    description=_("Zhipu proxy LLM configuration."),
    documentation_url="https://open.bigmodel.cn/dev/api/normal-model/glm-4#overview",
    show_in_ui=False,
)
@dataclass
class ZhipuDeployModelParameters(OpenAICompatibleDeployModelParameters):
    """Deploy model parameters for Zhipu."""

    provider: str = "proxy/zhipu"

    api_base: Optional[str] = field(
        default="${env:ZHIPUAI_BASE_URL:-https://open.bigmodel.cn/api/paas/v4}",
        metadata={
            "help": _("The base url of the Zhipu API."),
        },
    )

    api_key: Optional[str] = field(
        default="${env:ZHIPUAI_API_KEY}",
        metadata={
            "help": _("The API key of the Zhipu API."),
            "tags": "privacy",
        },
    )


def zhipu_generate_stream(
    model: ProxyModel, tokenizer, params, device, context_len=2048
):
    """Zhipu ai, see: https://open.bigmodel.cn/dev/api#overview"""
    model_params = model.get_params()
    print(f"Model: {model}, model_params: {model_params}")

    # TODO: Support convert_to_compatible_format config, zhipu not support system
    #  message
    # convert_to_compatible_format = params.get("convert_to_compatible_format", False)
    # history, systems = __convert_2_zhipu_messages(messages)
    client: ZhipuLLMClient = model.proxy_llm_client
    request = parse_model_request(params, client.default_model, stream=True)
    for r in client.sync_generate_stream(request):
        yield r


class ZhipuLLMClient(ProxyLLMClient):
    def __init__(
        self,
        model: Optional[str] = _DEFAULT_MODEL,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        model_alias: Optional[str] = _DEFAULT_MODEL,
        context_length: Optional[int] = 8192,
        executor: Optional[Executor] = None,
    ):
        try:
            from zhipuai import ZhipuAI

        except ImportError as exc:
            if (
                "No module named" in str(exc)
                or "cannot find module" in str(exc).lower()
            ):
                raise ValueError(
                    "The python package 'zhipuai' is not installed. "
                    "Please install it by running `pip install zhipuai`."
                ) from exc
            else:
                raise ValueError(
                    "Could not import python package: zhipuai "
                    "This may be due to a version that is too low. "
                    "Please upgrade the zhipuai package by running "
                    "`pip install --upgrade zhipuai`."
                ) from exc
        if not model:
            model = _DEFAULT_MODEL
        if not api_key:
            # Compatible with DB-GPT's config
            api_key = os.getenv("ZHIPU_PROXY_API_KEY")

        api_key = self._resolve_env_vars(api_key)
        api_base = self._resolve_env_vars(api_base)
        self._model = model
        self.client = ZhipuAI(api_key=api_key, base_url=api_base)

        super().__init__(
            model_names=[model, model_alias],
            context_length=context_length,
            executor=executor,
        )

    @classmethod
    def new_client(
        cls,
        model_params: ZhipuDeployModelParameters,
        default_executor: Optional[Executor] = None,
    ) -> "ZhipuLLMClient":
        return cls(
            model=model_params.real_provider_model_name,
            api_key=model_params.api_key,
            api_base=model_params.api_base,
            model_alias=model_params.real_provider_model_name,
            context_length=model_params.context_length,
            executor=default_executor,
        )

    @classmethod
    def param_class(cls) -> Type[LLMDeployModelParameters]:
        return ZhipuDeployModelParameters

    @classmethod
    def generate_stream_function(
        cls,
    ) -> Optional[Union[GenerateStreamFunction, AsyncGenerateStreamFunction]]:
        return zhipu_generate_stream

    @property
    def default_model(self) -> str:
        return self._model

    def sync_generate_stream(
        self,
        request: ModelRequest,
        message_converter: Optional[MessageConverter] = None,
    ) -> Iterator[ModelOutput]:
        request = self.local_covert_message(request, message_converter)

        messages = request.to_common_messages(support_system_role=False)

        model = request.model or self._model
        try:
            logger.debug(
                f"Send request to zhipu ai, model: {model}, request: {request}"
            )
            response = self.client.chat.completions.create(
                model=model,
                messages=messages,
                temperature=request.temperature,
                max_tokens=request.max_new_tokens,
                top_p=request.top_p,
                stream=True,
            )
            partial_text = ""
            for chunk in response:
                if not chunk.choices or not chunk.choices[0].delta:
                    continue
                delta_content = chunk.choices[0].delta.content
                finish_reason = chunk.choices[0].finish_reason
                partial_text += delta_content
                if logger.isEnabledFor(logging.DEBUG):
                    print(delta_content, end="")
                yield ModelOutput(
                    text=partial_text, error_code=0, finish_reason=finish_reason
                )
            if not partial_text:
                yield ModelOutput(text="**LLMServer Generate Empty.**", error_code=1)

        except Exception as e:
            yield ModelOutput(
                text=f"**LLMServer Generate Error, Please CheckErrorInfo.**: {e}",
                error_code=1,
            )


register_proxy_model_adapter(
    ZhipuLLMClient,
    supported_models=[
        ModelMetadata(
            model=["glm-4-plus", "glm-4-air", "glm-4-air-0111"],
            context_length=128 * 1024,
            max_output_length=4 * 1024,
            description="GLM-4 by Zhipu AI",
            link="https://bigmodel.cn/dev/howuse/model",
            function_calling=True,
        ),
        ModelMetadata(
            model=["glm-4-long"],
            context_length=1000 * 1024,
            max_output_length=4 * 1024,
            description="Long context GLM-4 by Zhipu AI",
            link="https://bigmodel.cn/dev/howuse/model",
            function_calling=True,
        ),
        ModelMetadata(
            model=["glm-4-flash", "glm-4-flashx"],
            context_length=128 * 1024,
            max_output_length=4 * 1024,
            description="Flash version of GLM-4 by Zhipu AI",
            link="https://bigmodel.cn/dev/howuse/model",
            function_calling=True,
        ),
        ModelMetadata(
            model=["glm-4-airx"],
            context_length=8 * 1024,
            max_output_length=4 * 1024,
            description="Quick response reasoning model by Zhipu AI",
            link="https://bigmodel.cn/dev/howuse/model",
            function_calling=True,
        ),
        ModelMetadata(
            model=["glm-zero-preview"],
            context_length=16 * 1024,
            max_output_length=12 * 1024,
            description="Reasoning model by Zhipu AI",
            link="https://bigmodel.cn/dev/howuse/model",
            function_calling=True,
        ),
    ],
)



import logging
from threading import Thread

import torch
from transformers import AutoModelForCausalLM, AutoTokenizer, TextIteratorStreamer

from dbgpt.core import ModelOutput

from ...utils.parse_utils import (
    _DEFAULT_THINK_START_TOKEN,
    ParsedChatMessage,
    parse_chat_message,
)

logger = logging.getLogger(__name__)


@torch.inference_mode()
def huggingface_chat_generate_stream(
    model: AutoModelForCausalLM,
    tokenizer: AutoTokenizer,
    params,
    device,
    context_len=4096,
):
    prompt = params["prompt"]
    temperature = float(params.get("temperature", 0.7))
    top_p = float(params.get("top_p", 1.0))
    echo = params.get("echo", False)
    max_new_tokens = int(params.get("max_new_tokens", 2048))
    stop_token_ids = params.get("stop_token_ids", [])
    do_sample = params.get("do_sample", True)
    custom_stop_words = params.get("custom_stop_words", [])
    think_start_token = params.get("think_start_token", _DEFAULT_THINK_START_TOKEN)
    is_reasoning_model = params.get("is_reasoning_model", False)

    input_ids = tokenizer(prompt).input_ids
    # input_ids = input_ids.to(device)
    if model.config.is_encoder_decoder:
        max_src_len = context_len
    else:  # truncate
        max_src_len = context_len - max_new_tokens - 1
    input_ids = input_ids[-max_src_len:]
    # input_echo_len = len(input_ids)
    input_ids = torch.as_tensor([input_ids], device=device)

    streamer = TextIteratorStreamer(
        tokenizer, skip_prompt=not echo, skip_special_tokens=True
    )

    base_kwargs = {
        "max_length": context_len,
        "temperature": temperature,
        "streamer": streamer,
        "top_p": top_p,
    }

    if stop_token_ids:
        base_kwargs["eos_token_id"] = stop_token_ids
    if do_sample is not None:
        base_kwargs["do_sample"] = do_sample

    logger.info(
        f"Predict with parameters: {base_kwargs}\ncustom_stop_words: "
        f"{custom_stop_words}"
    )

    generate_kwargs = {"input_ids": input_ids, **base_kwargs}
    thread = Thread(target=model.generate, kwargs=generate_kwargs)
    thread.start()
    text = ""
    usage = None
    msg = ParsedChatMessage()
    is_first = True
    for new_text in streamer:
        text += new_text
        if custom_stop_words:
            for stop_word in custom_stop_words:
                if text.endswith(stop_word):
                    text = text[: -len(stop_word)]

        if (
            prompt.rstrip().endswith(think_start_token)
            and is_reasoning_model
            and is_first
        ):
            text = think_start_token + "\n" + text
            is_first = False

        msg = parse_chat_message(text, extract_reasoning=is_reasoning_model)
        yield ModelOutput.build(
            msg.content,
            msg.reasoning_content,
            error_code=0,
            usage=usage,
            is_reasoning_model=is_reasoning_model,
        )



import json
import logging
from typing import Any, List, Optional, Tuple

from dbgpt.agent import (
    ActionOutput,
    Agent,
    AgentMemoryFragment,
    AgentMessage,
    ConversableAgent,
    ProfileConfig,
    ResourceType,
)
from dbgpt.agent.expand.actions.react_action import ReActAction
from dbgpt.core import ModelMessageRoleType
from dbgpt.util.configure import DynConfig
from dbgpt.util.json_utils import find_json_objects

logger = logging.getLogger(__name__)
_REACT_SYSTEM_TEMPLATE = """\
You are a {{ role }}, {% if name %}named {{ name }}.
{% endif %}your goal is {% if is_retry_chat %}{{ retry_goal }}
{% else %}{{ goal }}
{% endif %}.\
At the same time, please strictly abide by the constraints and specifications 
in the "IMPORTANT REMINDER" below.
{% if resource_prompt %}\
# ACTION SPACE #
{{ resource_prompt }} 
{% endif %}
{% if expand_prompt %}\
{{ expand_prompt }} 
{% endif %}\


# IMPORTANT REMINDER #
The current time is:{{now_time}}.
{% if constraints %}\
{% for constraint in constraints %}\
{{ loop.index }}. {{ constraint }}
{% endfor %}\
{% endif %}\


{% if is_retry_chat %}\
{% if retry_constraints %}\
{% for retry_constraint in retry_constraints %}\
{{ loop.index }}. {{ retry_constraint }}
{% endfor %}\
{% endif %}\
{% else %}\



{% endif %}\



{% if examples %}\
# EXAMPLE INTERACTION #
You can refer to the following examples:
{{ examples }}\
{% endif %}\

{% if most_recent_memories %}\
# History of Solving Task#
{{ most_recent_memories }}\
{% endif %}\

# RESPONSE FORMAT # 
{% if out_schema %} {{ out_schema }} {% endif %}\

################### TASK ###################
Please solve the task:
"""


_REACT_WRITE_MEMORY_TEMPLATE = """\
{% if question %}Question: {{ question }} {% endif %}
{% if assistant %}Assistant: {{ assistant }} {% endif %}
{% if observation %}Observation: {{ observation }} {% endif %}
"""


class ReActAgent(ConversableAgent):
    end_action_name: str = DynConfig(
        "terminate",
        category="agent",
        key="dbgpt_agent_expand_plugin_assistant_agent_end_action_name",
    )
    max_steps: int = DynConfig(
        10,
        category="agent",
        key="dbgpt_agent_expand_plugin_assistant_agent_max_steps",
    )
    profile: ProfileConfig = ProfileConfig(
        name=DynConfig(
            "ReAct",
            category="agent",
            key="dbgpt_agent_expand_plugin_assistant_agent_name",
        ),
        role=DynConfig(
            "ToolMaster",
            category="agent",
            key="dbgpt_agent_expand_plugin_assistant_agent_role",
        ),
        goal=DynConfig(
            "Read and understand the tool information given in the action space "
            "below to understand their capabilities and how to use them,and choosing "
            "the right tool to solve the task",
            category="agent",
            key="dbgpt_agent_expand_plugin_assistant_agent_goal",
        ),
        constraints=DynConfig(
            [
                "Achieve the goal step by step."
                "Each step, please read the parameter definition of the tool carefully "
                "and extract the specific parameters required to execute the tool "
                "from the user's goal.",
                "information in json format according to the following required format."
                "If there is an example, please refer to the sample format output.",
                "each step, you can only select one tool in action space.",
            ],
            category="agent",
            key="dbgpt_agent_expand_plugin_assistant_agent_constraints",
        ),
        system_prompt_template=_REACT_SYSTEM_TEMPLATE,
        write_memory_template=_REACT_WRITE_MEMORY_TEMPLATE,
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._init_actions([ReActAction])

    async def review(self, message: Optional[str], censored: Agent) -> Tuple[bool, Any]:
        """Review the message based on the censored message."""
        try:
            json_obj = find_json_objects(message)
            if len(json_obj) == 0:
                raise ValueError(
                    "No correct json object found in the message。"
                    "Please strictly output JSON in the defined "
                    "format, and only one action can be ouput each time. "
                )
            return True, json_obj[0]
        except Exception as e:
            logger.error(f"review error: {e}")
            raise e

    def validate_action(self, action_name: str) -> bool:
        tools = self.resource.get_resource_by_type(ResourceType.Tool)
        for tool in tools:
            if tool.name == action_name:
                return True
        raise ValueError(f"{action_name} is not in the action space.")

    async def generate_reply(
        self,
        received_message: AgentMessage,
        sender: Agent,
        reviewer: Optional[Agent] = None,
        rely_messages: Optional[List[AgentMessage]] = None,
        historical_dialogues: Optional[List[AgentMessage]] = None,
        is_retry_chat: bool = False,
        last_speaker_name: Optional[str] = None,
        **kwargs,
    ) -> AgentMessage:
        """Generate a reply based on the received messages."""
        try:
            logger.info(
                f"generate agent reply!sender={sender}, "
                f"rely_messages_len={rely_messages}"
            )
            self.validate_action(self.end_action_name)
            observation = AgentMessage(content="please start!")
            reply_message: AgentMessage = self._init_reply_message(
                received_message=received_message
            )
            thinking_messages, resource_info = await self._load_thinking_messages(
                received_message=observation,
                sender=sender,
                rely_messages=rely_messages,
                historical_dialogues=historical_dialogues,
                context=reply_message.get_dict_context(),
                is_retry_chat=is_retry_chat,
            )
            # attach current task to system prompt
            thinking_messages[0].content = (
                thinking_messages[0].content + "\n" + received_message.content
            )
            done = False
            max_steps = self.max_steps
            await self.write_memories(
                question=received_message.content,
                ai_message="",
            )
            while not done and max_steps > 0:
                ai_message = ""
                try:
                    # 1. thinking
                    llm_reply, model_name = await self.thinking(
                        thinking_messages, sender
                    )
                    reply_message.model_name = model_name
                    reply_message.resource_info = resource_info
                    ai_message = llm_reply
                    thinking_messages.append(
                        AgentMessage(role=ModelMessageRoleType.AI, content=llm_reply)
                    )
                    approve, json_obj = await self.review(llm_reply, self)
                    logger.info(f"jons_obj: {json_obj}")
                    action = json_obj["Action"]
                    thought = json_obj["Thought"]
                    action.update({"thought": thought})
                    reply_message.content = json.dumps(action, ensure_ascii=False)
                    tool_name = action["tool_name"]
                    self.validate_action(tool_name)
                    # 2. act
                    act_extent_param = self.prepare_act_param(
                        received_message=received_message,
                        sender=sender,
                        rely_messages=rely_messages,
                        historical_dialogues=historical_dialogues,
                    )
                    act_out: ActionOutput = await self.act(
                        message=reply_message,
                        sender=sender,
                        reviewer=reviewer,
                        is_retry_chat=is_retry_chat,
                        last_speaker_name=last_speaker_name,
                        **act_extent_param,
                    )
                    if act_out:
                        reply_message.action_report = act_out

                    # 3. obs
                    check_pass, reason = await self.verify(
                        reply_message, sender, reviewer
                    )
                    done = tool_name == self.end_action_name and check_pass
                    if check_pass:
                        logger.info(f"Observation:{act_out.content}")
                        thinking_messages.append(
                            AgentMessage(
                                role=ModelMessageRoleType.HUMAN,
                                content=f"Observation: {tool_name} "
                                f"output:{act_out.content}\n",
                            )
                        )
                        await self.write_memories(
                            question="",
                            ai_message=ai_message,
                            action_output=act_out,
                            check_pass=check_pass,
                        )
                    else:
                        observation = f"Observation: {reason}"
                        logger.info(f"Observation:{observation}")
                        thinking_messages.append(
                            AgentMessage(
                                role=ModelMessageRoleType.HUMAN, content=observation
                            )
                        )
                        await self.write_memories(
                            question="",
                            ai_message=ai_message,
                            check_pass=check_pass,
                            check_fail_reason=reason,
                        )
                    max_steps -= 1
                except Exception as e:
                    fail_reason = (
                        f"Observation: Exception occurs：({type(e).__name__}){e}."
                    )
                    logger.error(fail_reason)
                    thinking_messages.append(
                        AgentMessage(
                            role=ModelMessageRoleType.HUMAN, content=fail_reason
                        )
                    )
                    await self.write_memories(
                        question="",
                        ai_message=ai_message,
                        check_pass=False,
                        check_fail_reason=fail_reason,
                    )
            reply_message.success = done
            await self.adjust_final_message(True, reply_message)
            return reply_message
        except Exception as e:
            logger.exception("Generate reply exception!")
            err_message = AgentMessage(content=str(e))
            err_message.success = False
            return err_message

    async def write_memories(
        self,
        question: str,
        ai_message: str,
        action_output: Optional[ActionOutput] = None,
        check_pass: bool = True,
        check_fail_reason: Optional[str] = None,
    ) -> None:
        """Write the memories to the memory.

        We suggest you to override this method to save the conversation to memory
        according to your needs.

        Args:
            question(str): The question received.
            ai_message(str): The AI message, LLM output.
            action_output(ActionOutput): The action output.
            check_pass(bool): Whether the check pass.
            check_fail_reason(str): The check fail reason.
        """
        observation = ""
        if action_output and action_output.observations:
            observation = action_output.observations
        elif check_fail_reason:
            observation = check_fail_reason
        memory_map = {
            "question": question,
            "assistant": ai_message,
            "observation": observation,
        }
        write_memory_template = self.write_memory_template
        memory_content = self._render_template(write_memory_template, **memory_map)
        fragment = AgentMemoryFragment(memory_content)
        await self.memory.write(fragment)



"""The utility functions for chatting with the DAG task."""

import json
import traceback
from typing import Any, AsyncIterator, Dict, Optional

from ...interface.llm import ModelInferenceMetrics, ModelOutput
from ...schema.api import ChatCompletionResponseStreamChoice
from ..operators.base import BaseOperator
from ..trigger.http_trigger import CommonLLMHttpResponseBody


def is_chat_flow_type(output_obj: Any, is_class: bool = False) -> bool:
    """Check whether the output object is a chat flow type."""
    if is_class:
        return output_obj in (str, CommonLLMHttpResponseBody, ModelOutput)
    else:
        chat_types = (str, CommonLLMHttpResponseBody)
        return isinstance(output_obj, chat_types)


def is_agent_flow_type(output_obj: Any, is_class: bool = False) -> bool:
    """Check whether the output object is a agent flow type."""
    if is_class:
        return output_obj in (str, CommonLLMHttpResponseBody, ModelOutput)
    else:
        chat_types = (str, CommonLLMHttpResponseBody)
        return isinstance(output_obj, chat_types)


async def safe_chat_with_dag_task(
    task: BaseOperator, request: Any, covert_to_str: bool = False
) -> ModelOutput:
    """Chat with the DAG task.

    Args:
        task (BaseOperator): The DAG task to be executed.
        request (Any): The request to be passed to the DAG task.
        covert_to_str (bool, optional): Whether to convert the output to string.

    Returns:
        ModelOutput: The model output, the result is not incremental.
    """
    try:
        finish_reason = None
        usage = None
        metrics = None
        error_code = 0
        text = ""
        thinking_text = ""
        async for output in safe_chat_stream_with_dag_task(
            task, request, False, covert_to_str=covert_to_str
        ):
            finish_reason = output.finish_reason
            usage = output.usage
            metrics = output.metrics
            error_code = output.error_code
            if output.has_text:
                text = output.text
            if output.has_thinking:
                thinking_text = output.thinking_text
        return ModelOutput.build(
            text,
            thinking_text,
            error_code=error_code,
            metrics=metrics,
            usage=usage,
            finish_reason=finish_reason,
        )
    except Exception as e:
        return ModelOutput(error_code=1, text=str(e), incremental=False)


async def safe_chat_stream_with_dag_task(
    task: BaseOperator, request: Any, incremental: bool, covert_to_str: bool = False
) -> AsyncIterator[ModelOutput]:
    """Chat with the DAG task.

    This function is similar to `chat_stream_with_dag_task`, but it will catch the
    exception and return the error message.

    Args:
        task (BaseOperator): The DAG task to be executed.
        request (Any): The request to be passed to the DAG task.
        incremental (bool): Whether the output is incremental.
        covert_to_str (bool, optional): Whether to convert the output to string.

    Yields:
        ModelOutput: The model output.
    """
    try:
        async for output in chat_stream_with_dag_task(
            task, request, incremental, covert_to_str=covert_to_str
        ):
            yield output
    except Exception as e:
        simple_error_msg = str(e)
        if not simple_error_msg:
            simple_error_msg = traceback.format_exc()
        yield ModelOutput(error_code=1, text=simple_error_msg, incremental=incremental)
    finally:
        if task.streaming_operator and task.dag:
            await task.dag._after_dag_end(task.current_event_loop_task_id)


def _is_sse_output(task: BaseOperator) -> bool:
    """Check whether the DAG task is a server-sent event output.

    Args:
        task (BaseOperator): The DAG task.

    Returns:
        bool: Whether the DAG task is a server-sent event output.
    """
    return task.output_format is not None and task.output_format.upper() == "SSE"


async def chat_stream_with_dag_task(
    task: BaseOperator, request: Any, incremental: bool, covert_to_str: bool = False
) -> AsyncIterator[ModelOutput]:
    """Chat with the DAG task.

    Args:
        task (BaseOperator): The DAG task to be executed.
        request (Any): The request to be passed to the DAG task.
        incremental (bool): Whether the output is incremental.
        covert_to_str (bool, optional): Whether to convert the output to string.

    Yields:
        ModelOutput: The model output.
    """
    is_sse = _is_sse_output(task)
    if not task.streaming_operator:
        try:
            result = await task.call(request)
            model_output = parse_single_output(
                result, is_sse, covert_to_str=covert_to_str
            )
            model_output.incremental = incremental
            yield model_output
        except Exception as e:
            simple_error_msg = str(e)
            if not simple_error_msg:
                simple_error_msg = traceback.format_exc()
            yield ModelOutput(
                error_code=1, text=simple_error_msg, incremental=incremental
            )
    else:
        from dbgpt.model.utils.chatgpt_utils import OpenAIStreamingOutputOperator

        if OpenAIStreamingOutputOperator and isinstance(
            task, OpenAIStreamingOutputOperator
        ):
            full_text = ""
            full_thinking_text = ""
            async for output in await task.call_stream(request):
                model_output = parse_openai_output(output)
                # The output of the OpenAI streaming API is incremental
                if model_output.has_thinking:
                    full_thinking_text += model_output.thinking_text
                if model_output.has_text:
                    full_text += model_output.text
                model_output.incremental = incremental
                if not incremental:
                    model_output = ModelOutput.build(
                        full_text,
                        full_thinking_text,
                        error_code=model_output.error_code,
                        usage=model_output.usage,
                        finish_reason=model_output.finish_reason,
                    )
                yield model_output
                if not model_output.success:
                    break
        else:
            full_text = ""
            full_thinking_text = ""
            previous_text = ""
            previous_thinking_text = ""
            async for output in await task.call_stream(request):
                model_output = parse_single_output(
                    output, is_sse, covert_to_str=covert_to_str
                )
                model_output.incremental = incremental
                if task.incremental_output:
                    # Output is incremental, append the text
                    if model_output.has_thinking:
                        full_thinking_text += model_output.thinking_text
                    if model_output.has_text:
                        full_text += model_output.text
                else:
                    # Output is not incremental, last output is the full text
                    if model_output.has_thinking:
                        full_thinking_text = model_output.thinking_text

                    if model_output.has_text:
                        full_text = model_output.text
                if not incremental:
                    # Return the full text
                    model_output = ModelOutput.build(
                        full_text,
                        full_thinking_text,
                        error_code=model_output.error_code,
                        usage=model_output.usage,
                        finish_reason=model_output.finish_reason,
                    )
                else:
                    # Return the incremental text
                    delta_text = full_text[len(previous_text) :]
                    previous_text = (
                        full_text
                        if len(full_text) > len(previous_text)
                        else previous_text
                    )
                    delta_thinking_text = full_thinking_text[
                        len(previous_thinking_text) :
                    ]
                    previous_thinking_text = (
                        full_thinking_text
                        if len(full_thinking_text) > len(previous_thinking_text)
                        else previous_thinking_text
                    )
                    model_output = ModelOutput.build(
                        delta_text,
                        delta_thinking_text,
                        error_code=model_output.error_code,
                        usage=model_output.usage,
                        finish_reason=model_output.finish_reason,
                    )
                yield model_output
                if not model_output.success:
                    break


def parse_single_output(
    output: Any, is_sse: bool, covert_to_str: bool = False
) -> ModelOutput:
    """Parse the single output.

    Args:
        output (Any): The output to parse.
        is_sse (bool): Whether the output is in SSE format.
        covert_to_str (bool, optional): Whether to convert the output to string.
        Defaults to False.

    Returns:
        ModelOutput: The parsed output.
    """
    finish_reason: Optional[str] = None
    usage: Optional[Dict[str, Any]] = None
    metrics: Optional[ModelInferenceMetrics] = None

    if output is None:
        error_code = 1
        text = "The output is None!"
    elif isinstance(output, str):
        if is_sse:
            sse_output = parse_sse_data(output)
            if sse_output is None:
                error_code = 1
                text = "The output is not a SSE format"
            else:
                error_code = 0
                text = sse_output
        else:
            error_code = 0
            text = output
    elif isinstance(output, ModelOutput):
        # error_code = output.error_code
        # text = output.text
        # finish_reason = output.finish_reason
        # usage = output.usage
        # metrics = output.metrics
        return output
    elif isinstance(output, CommonLLMHttpResponseBody):
        error_code = output.error_code
        text = output.text
    elif isinstance(output, dict):
        error_code = 0
        text = json.dumps(output, ensure_ascii=False)
    elif covert_to_str:
        error_code = 0
        text = str(output)
    else:
        error_code = 1
        text = f"The output is not a valid format({type(output)})"
    return ModelOutput(
        error_code=error_code,
        text=text,
        finish_reason=finish_reason,
        usage=usage,
        metrics=metrics,
    )


def parse_openai_output(output: Any) -> ModelOutput:
    """Parse the OpenAI output.

    Args:
        output (Any): The output to parse. It must be a stream format.

    Returns:
        ModelOutput: The parsed output.
    """
    text = ""
    if not isinstance(output, str):
        return ModelOutput(
            error_code=1,
            text="The output is not a stream format",
        )
    if output.strip() == "data: [DONE]" or output.strip() == "data:[DONE]":
        return ModelOutput(error_code=0, text="")
    if not output.startswith("data:"):
        return ModelOutput(
            error_code=1,
            text="The output is not a stream format",
        )

    sse_output = parse_sse_data(output)
    if sse_output is None:
        return ModelOutput(error_code=1, text="The output is not a SSE format")
    json_data = sse_output.strip()
    try:
        dict_data = json.loads(json_data)
    except Exception as e:
        return ModelOutput(
            error_code=1,
            text=f"Invalid JSON data: {json_data}, {e}",
        )
    if "choices" not in dict_data:
        return ModelOutput(
            error_code=1,
            text=dict_data.get("text", "Unknown error"),
        )
    choices = dict_data["choices"]
    finish_reason: Optional[str] = None
    reasoning_content: Optional[str] = None
    if choices:
        choice = choices[0]
        delta_data = ChatCompletionResponseStreamChoice(**choice)
        if delta_data.delta.content:
            text = delta_data.delta.content
        if delta_data.delta.reasoning_content:
            reasoning_content = delta_data.delta.reasoning_content
        finish_reason = delta_data.finish_reason
    return ModelOutput.build(text, reasoning_content, finish_reason=finish_reason)


def parse_sse_data(output: str) -> Optional[str]:
    r"""Parse the SSE data.

    Just keep the data part.

    Examples:
        .. code-block:: python

            from dbgpt.core.awel.util.chat_util import parse_sse_data

            assert parse_sse_data("data: [DONE]") == "[DONE]"
            assert parse_sse_data("data:[DONE]") == "[DONE]"
            assert parse_sse_data("data: Hello") == "Hello"
            assert parse_sse_data("data: Hello\n") == "Hello"
            assert parse_sse_data("data: Hello\r\n") == "Hello"
            assert parse_sse_data("data: Hi, what's up?") == "Hi, what's up?"

    Args:
        output (str): The output.

    Returns:
        Optional[str]: The parsed data.
    """
    if output.startswith("data:"):
        output = output.strip()
        if output.startswith("data: "):
            output = output[6:]
        else:
            output = output[5:]

        return output
    else:
        return None



"""Resources for the agent."""

import dataclasses
import json
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Generic, List, Optional, Tuple, Type, TypeVar, cast

from pydantic import field_validator

from dbgpt._private.pydantic import BaseModel, model_to_dict
from dbgpt.core import Chunk
from dbgpt.util.parameter_utils import BaseParameters, _get_parameter_descriptions

P = TypeVar("P", bound="ResourceParameters")
T = TypeVar("T", bound="Resource")


class ResourceType(str, Enum):
    """Resource type enumeration."""

    DB = "database"
    Knowledge = "knowledge"
    Internet = "internet"
    Tool = "tool"
    Plugin = "plugin"
    TextFile = "text_file"
    ExcelFile = "excel_file"
    ImageFile = "image_file"
    AWELFlow = "awel_flow"
    App = "app"
    # Resource type for resource pack
    Pack = "pack"


@dataclasses.dataclass
class ResourceParameters(BaseParameters):
    """Resource parameters class.

    It defines the parameters for building a resource.
    """

    name: str = dataclasses.field(metadata={"help": "Resource name", "tags": "fixed"})

    @classmethod
    def _resource_version(cls) -> str:
        """Return the resource version."""
        return "v2"

    @classmethod
    def to_configurations(
        cls,
        parameters: Type["ResourceParameters"],
        version: Optional[str] = None,
    ) -> Any:
        """Convert the parameters to configurations."""
        return _get_parameter_descriptions(parameters)


class Resource(ABC, Generic[P]):
    """Resource for the agent."""

    @classmethod
    @abstractmethod
    def type(cls) -> ResourceType:
        """Return the resource type."""

    @classmethod
    def type_alias(cls) -> str:
        """Return the resource type alias."""
        return cls.type().value

    @property
    @abstractmethod
    def name(self) -> str:
        """Return the resource name."""

    @classmethod
    def resource_parameters_class(cls, **kwargs) -> Type[P]:
        """Return the parameters class."""
        return ResourceParameters

    def prefer_resource_parameters_class(self, **kwargs) -> Type[P]:
        """Return the parameters class.

        You can override this method to return a different parameters class.
        It will be used to initialize the resource with parameters.
        """
        return self.resource_parameters_class(**kwargs)

    def initialize_with_parameters(self, resource_parameters: P):
        """Initialize the resource with parameters."""
        pass

    def preload_resource(self):
        """Preload the resource."""
        pass

    @classmethod
    def from_resource(
        cls: Type[T],
        resource: Optional["Resource"],
        expected_type: Optional[ResourceType] = None,
    ) -> List[T]:
        """Create a resource from another resource.

        Another resource can be a pack or a single resource, if it is a pack, it will
        return all resources which type is the same as the current resource.

        Args:
            resource(Resource): The resource.
            expected_type(ResourceType): The expected resource type.
        Returns:
            List[Resource]: The resources.
        """
        if not resource:
            return []
        typed_resources = []
        for r in resource.get_resource_by_type(expected_type or cls.type()):
            typed_resources.append(cast(T, r))
        return typed_resources

    async def get_resources_info(
        self,
        *,
        lang: str = "en",
        prompt_type: str = "default",
        question: Optional[str] = None,
        resources: Optional[List] = None,
        **kwargs,
    ):
        """Get prompts for multiple resources at the same time.

        Args:
            lang(str): The language.
            prompt_type(str): The prompt type.
            question(str): The question.
            resource_name(str): The resource name, just for the pack, it will be used
                to select specific resource in the pack.
        """

    @abstractmethod
    async def get_prompt(
        self,
        *,
        lang: str = "en",
        prompt_type: str = "default",
        question: Optional[str] = None,
        resource_name: Optional[str] = None,
        **kwargs,
    ) -> Tuple[str, Optional[Dict]]:
        """Get the prompt.

        Args:
            lang(str): The language.
            prompt_type(str): The prompt type.
            question(str): The question.
            resource_name(str): The resource name, just for the pack, it will be used
                to select specific resource in the pack.
        """

    async def get_resources(
        self,
        lang: str = "en",
        prompt_type: str = "default",
        question: Optional[str] = None,
        resource_name: Optional[str] = None,
    ) -> Tuple[Optional[List[Chunk]], str, Optional[Dict]]:
        """Get the resources."""
        raise NotImplementedError

    def execute(self, *args, resource_name: Optional[str] = None, **kwargs) -> Any:
        """Execute the resource."""
        raise NotImplementedError

    async def async_execute(
        self, *args, resource_name: Optional[str] = None, **kwargs
    ) -> Any:
        """Execute the resource asynchronously."""
        raise NotImplementedError

    @property
    def is_async(self) -> bool:
        """Return whether the resource is asynchronous."""
        return False

    @property
    def is_pack(self) -> bool:
        """Return whether the resource is a pack."""
        return False

    @property
    def sub_resources(self) -> List["Resource"]:
        """Return the resources."""
        if not self.is_pack:
            raise ValueError("The resource is not a pack, no sub-resources.")
        return []

    def get_resource_by_type(self, resource_type: ResourceType) -> List["Resource"]:
        """Get resources by type.

        If the resource is a pack, it will search the sub-resources. Otherwise, it will
        return itself if the type matches.

        Args:
            resource_type(ResourceType): The resource type.

        Returns:
            List[Resource]: The resources.
        """
        if not self.is_pack:
            if self.type() == resource_type:
                return [self]
            else:
                return []
        resources = []
        for resource in self.sub_resources:
            if resource.type() == resource_type:
                resources.append(resource)
        return resources


class AgentResource(BaseModel):
    """Agent resource class."""

    type: str
    value: str
    name: Optional[str] = None

    is_dynamic: bool = (
        False  # Is the current resource predefined or dynamically passed in?
    )
    context: Optional[dict] = None

    def resource_prompt_template(self, **kwargs) -> str:
        """Get the resource prompt template."""
        return "{data_type}  --{data_introduce}"

    @field_validator("value", mode="before")
    def parse_value(cls, value):
        """Parse value."""
        return str(value)

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> Optional["AgentResource"]:
        """Create an AgentResource object from a dictionary."""
        if d is None:
            return None
        return AgentResource(
            type=d.get("type"),
            name=d.get("name"),
            introduce=d.get("introduce"),
            value=d.get("value", None),
            is_dynamic=d.get("is_dynamic", False),
            context=d.get("context", None),
        )

    @staticmethod
    def from_json_list_str(d: Optional[str]) -> Optional[List["AgentResource"]]:
        """Create a list of AgentResource objects from a json string."""
        if d is None:
            return None
        try:
            json_array = json.loads(d)
        except Exception:
            raise ValueError(f"Illegal AgentResource json string！{d}")
        if not isinstance(json_array, list):
            raise ValueError(f"Illegal AgentResource json string！{d}")
        json_list = []
        for item in json_array:
            r = AgentResource.from_dict(item)
            if r:
                json_list.append(r)
        return json_list

    def to_dict(self) -> Dict[str, Any]:
        """Convert the AgentResource object to a dictionary."""
        temp = model_to_dict(self)
        for field, value in temp.items():
            if isinstance(value, Enum):
                temp[field] = value.value
        return temp



import json
import logging
from typing import Optional

from dbgpt.agent import ResourceType
from dbgpt.agent.expand.actions.tool_action import ToolAction, ToolInput
from dbgpt.vis import Vis, VisPlugin

logger = logging.getLogger(__name__)


class ReActAction(ToolAction):
    """ReAct action class."""

    def __init__(self, **kwargs):
        """Tool action init."""
        super().__init__(**kwargs)
        self._render_protocol = VisPlugin()

    @property
    def resource_need(self) -> Optional[ResourceType]:
        """Return the resource type needed for the action."""
        return ResourceType.Tool

    @property
    def render_protocol(self) -> Optional[Vis]:
        """Return the render protocol."""
        return self._render_protocol

    @property
    def out_model_type(self):
        """Return the output model type."""
        return ToolInput

    @property
    def ai_out_schema(self) -> Optional[str]:
        """Return the AI output schema."""
        out_put_schema = {
            "Thought": "Summary of thoughts to the user",
            "Action": {
                "tool_name": "The name of a tool that can be used to answer "
                "the current"
                "question or solve the current task.",
                "args": {
                    "arg name1": "arg value1",
                    "arg name2": "arg value2",
                },
            },
        }

        return f"""Please response in the following json format:
        {json.dumps(out_put_schema, indent=2, ensure_ascii=False)}
        Make sure the response is correct json and can be parsed by Python json.loads.
        """
